// --------------------------------------------------------------------
// Low accuracy FP for DSP
// {sign, exp, mantissa} 
// sign is zero if positive, 
// exp is 8 bit offset binary (e.g. 2**0 is 128 and 2**1 is 129
// mantissa is 9 bit fraction >=0.5 and <1.0 OR ZERO

// --------------------------------------------------------------------
//
// Bruce Land, August 2008, Cornell University
//
// --------------------------------------------------------------------
////////////////////////////////////////////////////////////////////////

module DE2_TOP
	(
		////////////////////	Clock Input	 	////////////////////	 
		CLOCK_27,						//	27 MHz
		CLOCK_50,						//	50 MHz
		EXT_CLOCK,						//	External Clock
		////////////////////	Push Button		////////////////////
		KEY,							//	Pushbutton[3:0]
		////////////////////	DPDT Switch		////////////////////
		SW,								//	Toggle Switch[17:0]
		////////////////////	7-SEG Dispaly	////////////////////
		HEX0,							//	Seven Segment Digit 0
		HEX1,							//	Seven Segment Digit 1
		HEX2,							//	Seven Segment Digit 2
		HEX3,							//	Seven Segment Digit 3
		HEX4,							//	Seven Segment Digit 4
		HEX5,							//	Seven Segment Digit 5
		HEX6,							//	Seven Segment Digit 6
		HEX7,							//	Seven Segment Digit 7
		////////////////////////	LED		////////////////////////
		LEDG,							//	LED Green[8:0]
		LEDR,							//	LED Red[17:0]
		////////////////////////	UART	////////////////////////
		UART_TXD,						//	UART Transmitter
		UART_RXD,						//	UART Receiver
		////////////////////////	IRDA	////////////////////////
		IRDA_TXD,						//	IRDA Transmitter
		IRDA_RXD,						//	IRDA Receiver
		/////////////////////	SDRAM Interface		////////////////
		DRAM_DQ,						//	SDRAM Data bus 16 Bits
		DRAM_ADDR,						//	SDRAM Address bus 12 Bits
		DRAM_LDQM,						//	SDRAM Low-byte Data Mask 
		DRAM_UDQM,						//	SDRAM High-byte Data Mask
		DRAM_WE_N,						//	SDRAM Write Enable
		DRAM_CAS_N,						//	SDRAM Column Address Strobe
		DRAM_RAS_N,						//	SDRAM Row Address Strobe
		DRAM_CS_N,						//	SDRAM Chip Select
		DRAM_BA_0,						//	SDRAM Bank Address 0
		DRAM_BA_1,						//	SDRAM Bank Address 0
		DRAM_CLK,						//	SDRAM Clock
		DRAM_CKE,						//	SDRAM Clock Enable
		////////////////////	Flash Interface		////////////////
		FL_DQ,							//	FLASH Data bus 8 Bits
		FL_ADDR,						//	FLASH Address bus 22 Bits
		FL_WE_N,						//	FLASH Write Enable
		FL_RST_N,						//	FLASH Reset
		FL_OE_N,						//	FLASH Output Enable
		FL_CE_N,						//	FLASH Chip Enable
		////////////////////	SRAM Interface		////////////////
		SRAM_DQ,						//	SRAM Data bus 16 Bits
		SRAM_ADDR,						//	SRAM Address bus 18 Bits
		SRAM_UB_N,						//	SRAM High-byte Data Mask 
		SRAM_LB_N,						//	SRAM Low-byte Data Mask 
		SRAM_WE_N,						//	SRAM Write Enable
		SRAM_CE_N,						//	SRAM Chip Enable
		SRAM_OE_N,						//	SRAM Output Enable
		////////////////////	ISP1362 Interface	////////////////
		OTG_DATA,						//	ISP1362 Data bus 16 Bits
		OTG_ADDR,						//	ISP1362 Address 2 Bits
		OTG_CS_N,						//	ISP1362 Chip Select
		OTG_RD_N,						//	ISP1362 Write
		OTG_WR_N,						//	ISP1362 Read
		OTG_RST_N,						//	ISP1362 Reset
		OTG_FSPEED,						//	USB Full Speed,	0 = Enable, Z = Disable
		OTG_LSPEED,						//	USB Low Speed, 	0 = Enable, Z = Disable
		OTG_INT0,						//	ISP1362 Interrupt 0
		OTG_INT1,						//	ISP1362 Interrupt 1
		OTG_DREQ0,						//	ISP1362 DMA Request 0
		OTG_DREQ1,						//	ISP1362 DMA Request 1
		OTG_DACK0_N,					//	ISP1362 DMA Acknowledge 0
		OTG_DACK1_N,					//	ISP1362 DMA Acknowledge 1
		////////////////////	LCD Module 16X2		////////////////
		LCD_ON,							//	LCD Power ON/OFF
		LCD_BLON,						//	LCD Back Light ON/OFF
		LCD_RW,							//	LCD Read/Write Select, 0 = Write, 1 = Read
		LCD_EN,							//	LCD Enable
		LCD_RS,							//	LCD Command/Data Select, 0 = Command, 1 = Data
		LCD_DATA,						//	LCD Data bus 8 bits
		////////////////////	SD_Card Interface	////////////////
		SD_DAT,							//	SD Card Data
		SD_DAT3,						//	SD Card Data 3
		SD_CMD,							//	SD Card Command Signal
		SD_CLK,							//	SD Card Clock
		////////////////////	USB JTAG link	////////////////////
		TDI,  							// CPLD -> FPGA (data in)
		TCK,  							// CPLD -> FPGA (clk)
		TCS,  							// CPLD -> FPGA (CS)
	    TDO,  							// FPGA -> CPLD (data out)
		////////////////////	I2C		////////////////////////////
		I2C_SDAT,						//	I2C Data
		I2C_SCLK,						//	I2C Clock
		////////////////////	PS2		////////////////////////////
		PS2_DAT,						//	PS2 Data
		PS2_CLK,						//	PS2 Clock
		////////////////////	VGA		////////////////////////////
		VGA_CLK,   						//	VGA Clock
		VGA_HS,							//	VGA H_SYNC
		VGA_VS,							//	VGA V_SYNC
		VGA_BLANK,						//	VGA BLANK
		VGA_SYNC,						//	VGA SYNC
		VGA_R,   						//	VGA Red[9:0]
		VGA_G,	 						//	VGA Green[9:0]
		VGA_B,  						//	VGA Blue[9:0]
		////////////	Ethernet Interface	////////////////////////
		ENET_DATA,						//	DM9000A DATA bus 16Bits
		ENET_CMD,						//	DM9000A Command/Data Select, 0 = Command, 1 = Data
		ENET_CS_N,						//	DM9000A Chip Select
		ENET_WR_N,						//	DM9000A Write
		ENET_RD_N,						//	DM9000A Read
		ENET_RST_N,						//	DM9000A Reset
		ENET_INT,						//	DM9000A Interrupt
		ENET_CLK,						//	DM9000A Clock 25 MHz
		////////////////	Audio CODEC		////////////////////////
		AUD_ADCLRCK,					//	Audio CODEC ADC LR Clock
		AUD_ADCDAT,						//	Audio CODEC ADC Data
		AUD_DACLRCK,					//	Audio CODEC DAC LR Clock
		AUD_DACDAT,						//	Audio CODEC DAC Data
		AUD_BCLK,						//	Audio CODEC Bit-Stream Clock
		AUD_XCK,						//	Audio CODEC Chip Clock
		////////////////	TV Decoder		////////////////////////
		TD_DATA,    					//	TV Decoder Data bus 8 bits
		TD_HS,							//	TV Decoder H_SYNC
		TD_VS,							//	TV Decoder V_SYNC
		TD_RESET,						//	TV Decoder Reset
		////////////////////	GPIO	////////////////////////////
		GPIO_0,							//	GPIO Connection 0
		GPIO_1							//	GPIO Connection 1
	);

////////////////////////	Clock Input	 	////////////////////////
input			CLOCK_27;				//	27 MHz
input			CLOCK_50;				//	50 MHz
input			EXT_CLOCK;				//	External Clock
////////////////////////	Push Button		////////////////////////
input	[3:0]	KEY;					//	Pushbutton[3:0]
////////////////////////	DPDT Switch		////////////////////////
input	[17:0]	SW;						//	Toggle Switch[17:0]
////////////////////////	7-SEG Dispaly	////////////////////////
output	[6:0]	HEX0;					//	Seven Segment Digit 0
output	[6:0]	HEX1;					//	Seven Segment Digit 1
output	[6:0]	HEX2;					//	Seven Segment Digit 2
output	[6:0]	HEX3;					//	Seven Segment Digit 3
output	[6:0]	HEX4;					//	Seven Segment Digit 4
output	[6:0]	HEX5;					//	Seven Segment Digit 5
output	[6:0]	HEX6;					//	Seven Segment Digit 6
output	[6:0]	HEX7;					//	Seven Segment Digit 7
////////////////////////////	LED		////////////////////////////
output	[8:0]	LEDG;					//	LED Green[8:0]
output	[17:0]	LEDR;					//	LED Red[17:0]
////////////////////////////	UART	////////////////////////////
output			UART_TXD;				//	UART Transmitter
input			UART_RXD;				//	UART Receiver
////////////////////////////	IRDA	////////////////////////////
output			IRDA_TXD;				//	IRDA Transmitter
input			IRDA_RXD;				//	IRDA Receiver
///////////////////////		SDRAM Interface	////////////////////////
inout	[15:0]	DRAM_DQ;				//	SDRAM Data bus 16 Bits
output	[11:0]	DRAM_ADDR;				//	SDRAM Address bus 12 Bits
output			DRAM_LDQM;				//	SDRAM Low-byte Data Mask 
output			DRAM_UDQM;				//	SDRAM High-byte Data Mask
output			DRAM_WE_N;				//	SDRAM Write Enable
output			DRAM_CAS_N;				//	SDRAM Column Address Strobe
output			DRAM_RAS_N;				//	SDRAM Row Address Strobe
output			DRAM_CS_N;				//	SDRAM Chip Select
output			DRAM_BA_0;				//	SDRAM Bank Address 0
output			DRAM_BA_1;				//	SDRAM Bank Address 0
output			DRAM_CLK;				//	SDRAM Clock
output			DRAM_CKE;				//	SDRAM Clock Enable
////////////////////////	Flash Interface	////////////////////////
inout	[7:0]	FL_DQ;					//	FLASH Data bus 8 Bits
output	[21:0]	FL_ADDR;				//	FLASH Address bus 22 Bits
output			FL_WE_N;				//	FLASH Write Enable
output			FL_RST_N;				//	FLASH Reset
output			FL_OE_N;				//	FLASH Output Enable
output			FL_CE_N;				//	FLASH Chip Enable
////////////////////////	SRAM Interface	////////////////////////
inout	[15:0]	SRAM_DQ;				//	SRAM Data bus 16 Bits
output	[17:0]	SRAM_ADDR;				//	SRAM Address bus 18 Bits
output			SRAM_UB_N;				//	SRAM High-byte Data Mask 
output			SRAM_LB_N;				//	SRAM Low-byte Data Mask 
output			SRAM_WE_N;				//	SRAM Write Enable
output			SRAM_CE_N;				//	SRAM Chip Enable
output			SRAM_OE_N;				//	SRAM Output Enable
////////////////////	ISP1362 Interface	////////////////////////
inout	[15:0]	OTG_DATA;				//	ISP1362 Data bus 16 Bits
output	[1:0]	OTG_ADDR;				//	ISP1362 Address 2 Bits
output			OTG_CS_N;				//	ISP1362 Chip Select
output			OTG_RD_N;				//	ISP1362 Write
output			OTG_WR_N;				//	ISP1362 Read
output			OTG_RST_N;				//	ISP1362 Reset
output			OTG_FSPEED;				//	USB Full Speed,	0 = Enable, Z = Disable
output			OTG_LSPEED;				//	USB Low Speed, 	0 = Enable, Z = Disable
input			OTG_INT0;				//	ISP1362 Interrupt 0
input			OTG_INT1;				//	ISP1362 Interrupt 1
input			OTG_DREQ0;				//	ISP1362 DMA Request 0
input			OTG_DREQ1;				//	ISP1362 DMA Request 1
output			OTG_DACK0_N;			//	ISP1362 DMA Acknowledge 0
output			OTG_DACK1_N;			//	ISP1362 DMA Acknowledge 1
////////////////////	LCD Module 16X2	////////////////////////////
inout	[7:0]	LCD_DATA;				//	LCD Data bus 8 bits
output			LCD_ON;					//	LCD Power ON/OFF
output			LCD_BLON;				//	LCD Back Light ON/OFF
output			LCD_RW;					//	LCD Read/Write Select, 0 = Write, 1 = Read
output			LCD_EN;					//	LCD Enable
output			LCD_RS;					//	LCD Command/Data Select, 0 = Command, 1 = Data
////////////////////	SD Card Interface	////////////////////////
inout			SD_DAT;					//	SD Card Data
inout			SD_DAT3;				//	SD Card Data 3
inout			SD_CMD;					//	SD Card Command Signal
output			SD_CLK;					//	SD Card Clock
////////////////////////	I2C		////////////////////////////////
inout			I2C_SDAT;				//	I2C Data
output			I2C_SCLK;				//	I2C Clock
////////////////////////	PS2		////////////////////////////////
input		 	PS2_DAT;				//	PS2 Data
input			PS2_CLK;				//	PS2 Clock
////////////////////	USB JTAG link	////////////////////////////
input  			TDI;					// CPLD -> FPGA (data in)
input  			TCK;					// CPLD -> FPGA (clk)
input  			TCS;					// CPLD -> FPGA (CS)
output 			TDO;					// FPGA -> CPLD (data out)
////////////////////////	VGA			////////////////////////////
output			VGA_CLK;   				//	VGA Clock
output			VGA_HS;					//	VGA H_SYNC
output			VGA_VS;					//	VGA V_SYNC
output			VGA_BLANK;				//	VGA BLANK
output			VGA_SYNC;				//	VGA SYNC
output	[9:0]	VGA_R;   				//	VGA Red[9:0]
output	[9:0]	VGA_G;	 				//	VGA Green[9:0]
output	[9:0]	VGA_B;   				//	VGA Blue[9:0]
////////////////	Ethernet Interface	////////////////////////////
inout	[15:0]	ENET_DATA;				//	DM9000A DATA bus 16Bits
output			ENET_CMD;				//	DM9000A Command/Data Select, 0 = Command, 1 = Data
output			ENET_CS_N;				//	DM9000A Chip Select
output			ENET_WR_N;				//	DM9000A Write
output			ENET_RD_N;				//	DM9000A Read
output			ENET_RST_N;				//	DM9000A Reset
input			ENET_INT;				//	DM9000A Interrupt
output			ENET_CLK;				//	DM9000A Clock 25 MHz
////////////////////	Audio CODEC		////////////////////////////
inout			AUD_ADCLRCK;			//	Audio CODEC ADC LR Clock
input			AUD_ADCDAT;				//	Audio CODEC ADC Data
inout			AUD_DACLRCK;			//	Audio CODEC DAC LR Clock
output			AUD_DACDAT;				//	Audio CODEC DAC Data
inout			AUD_BCLK;				//	Audio CODEC Bit-Stream Clock
output			AUD_XCK;				//	Audio CODEC Chip Clock
////////////////////	TV Devoder		////////////////////////////
input	[7:0]	TD_DATA;    			//	TV Decoder Data bus 8 bits
input			TD_HS;					//	TV Decoder H_SYNC
input			TD_VS;					//	TV Decoder V_SYNC
output			TD_RESET;				//	TV Decoder Reset
////////////////////////	GPIO	////////////////////////////////
inout	[35:0]	GPIO_0;					//	GPIO Connection 0
inout	[35:0]	GPIO_1;					//	GPIO Connection 1

assign	LCD_ON		=	1'b0;
assign	LCD_BLON	=	1'b0;

//	All inout port turn to tri-state
assign	DRAM_DQ		=	16'hzzzz;
assign	FL_DQ		=	8'hzz;
assign	SRAM_DQ		=	16'hzzzz;
assign	OTG_DATA	=	16'hzzzz;
assign	LCD_DATA	=	8'hzz;
assign	SD_DAT		=	1'bz;
assign	I2C_SDAT	=	1'bz;
assign	ENET_DATA	=	16'hzzzz;
assign	AUD_ADCLRCK	=	1'bz;
assign	AUD_DACLRCK	=	1'bz;
assign	AUD_BCLK	=	1'bz;
assign	GPIO_0		=	36'hzzzzzzzzz;
assign	GPIO_1		=	36'hzzzzzzzzz;

//////////////////////
	
	// set f1=2 -- sign=0, exp=2+128 (128 offset), mantissa=h100 (or 1/2)
	// assign f1 = {1'b0, 8'd130, 9'h100} ;
	// set f1=0.75 -- sign=0, exp=128 (128 offset), mantissa=h180 (or 3/4)
	//assign f1 = {1'b0, 8'd128, 8'hc0, 1'b0} ;
	
	// set the gain as a fp number
	wire [17:0] gain ;
	assign gain = SW ;
	
	// fout mantissa
	HexDigit H0(HEX0,gain[0]);
	HexDigit H1(HEX1,gain[4:1]);
	HexDigit H2(HEX2,gain[8:5]);
	HexDigit H3(HEX3,4'b0); // turn it off
	// fout exp
	HexDigit H4(HEX4,gain[12:9]);
	HexDigit H5(HEX5,gain[16:13]);
	// fout sign
	HexDigit H6(HEX6,gain[17]);
	HexDigit H7(HEX7,4'b0); // turn it off
	
	//make a direct digital systnesis accumulator 
	// and output the top 8 bits to connector JP1
	// and output a 10-bit sine wave to the VGA plug
	//DDS hardware
	reg [31:0] DDS_accum1, DDS_accum2 ;
	always @ (posedge CLOCK_50) 
	begin
		DDS_accum1 <= DDS_accum1 + 32'h00010000 ;
		DDS_accum2 <= DDS_accum2 + 32'h00001000 ;
	end
	
	
	//hook up the ROM table
	sync_rom sineTable1(CLOCK_50, DDS_accum1[31:24], sine_out1);
	sync_rom sineTable2(CLOCK_50, DDS_accum2[31:24], sine_out2);
	
	// convert to sines to FP
	wire [17:0] fp_sine1, fp_sine2, fp_mod ;
	wire [9:0] sine_out1, sine_out2, sine_mod ; 
	int2fp i2f1(fp_sine1, sin_out1);
	int2fp i2f2(fp_sine2, sin_out2);
	fpmult m1(fp_mod, fp_sine1, fp_sine2);
	fp2int f2i1(sin_mod, fp_mod);
	
	// then multiply
	// the convert back to int
	
	//use the VGA DAC for an audio signal 
	assign VGA_R = sine_out1 ; //10-bits
	assign VGA_G = sine_out2 ; //10-bits
	assign VGA_B = sine_mod ; //10-bits
	//assign VGA_R = sine_out1 ; //10-bits
	assign VGA_SYNC = 1 ;
	assign VGA_BLANK = 1 ;
	assign VGA_CLK = CLOCK_50 ;
endmodule 

//////////////////////////////////////////////
// Decode one hex digit for LED 7-seg display
//////////////////////////////////////////////////
module HexDigit(segs, num);
	input [3:0] num	;		//the hex digit to be displayed
	output [6:0] segs ;		//actual LED segments
	reg [6:0] segs ;
	always @ (num)
	begin
		case (num)
				4'h0: segs = 7'b1000000;
				4'h1: segs = 7'b1111001;
				4'h2: segs = 7'b0100100;
				4'h3: segs = 7'b0110000;
				4'h4: segs = 7'b0011001;
				4'h5: segs = 7'b0010010;
				4'h6: segs = 7'b0000010;
				4'h7: segs = 7'b1111000;
				4'h8: segs = 7'b0000000;
				4'h9: segs = 7'b0010000;
				4'ha: segs = 7'b0001000;
				4'hb: segs = 7'b0000011;
				4'hc: segs = 7'b1000110;
				4'hd: segs = 7'b0100001;
				4'he: segs = 7'b0000110;
				4'hf: segs = 7'b0001110;
				default segs = 7'b1111111;
		endcase
	end
endmodule

///////////////////////////////////////////////////////////
// 10-bit int to float
///////////////////////////////////////////////////////////
// The 50 MHz VGA DAC uses 10-bit binary with 
// zero volts mapped to digital zero
// and about 1.5 volts mapped to digital 1024
// For the tables below, the 'zero' of the sine wave
// is mapped to 10'h100
///////////////////////////////////////////////////////////
module int2fp (fp_out, int_in);
	input [9:0] int_in ;
	output [17:0] fp_out ;
	
	wire sign ;
	reg [7:0] exponent ;
	reg [8:0] mantissa, mtemp ;
	
	always @(*)
	begin
		// get mantissa magnitude and truncate to 9 bits 
		// (top bit of int_in is zero)
		if (int_in >= 10'h100) 
			mtemp = int_in[8:0] - 9'h100 ;
		else 
			mtemp = 9'h100 - int_in[8:0] ;  
		// normalize mtemp 
		casex (mtemp)
			9'b1xxxxxxxx: 
			begin 
				mantissa = mtemp ;
				exponent = 8'h80 ;
			end 
			9'b01xxxxxxx: 
			begin 
				mantissa = mtemp<<1 ;
				exponent = 8'h80-1 ;
			end 
			9'b001xxxxxx: 
			begin 
				mantissa = mtemp<<2 ;
				exponent = 8'h80-2 ;
			end 
			9'b0001xxxxx: 
			begin 
				mantissa = mtemp<<3 ;
				exponent = 8'h80-3 ;
			end 
			9'b00001xxxx: 
			begin 
				mantissa = mtemp<<4 ;
				exponent = 8'h80-4 ;
			end 
			9'b000001xxx: 
			begin 
				mantissa = mtemp<<5 ;
				exponent = 8'h80-5 ;
			end 
			9'b0000001xx:
			begin 
				mantissa = mtemp<<6 ;
				exponent = 8'h80-6 ;
			end 
			9'b00000001x:
			begin 
				mantissa = mtemp<<7 ;
				exponent = 8'h80-7 ;
			end 
			9'b000000001:
			begin 
				mantissa = mtemp<<8 ;
				exponent = 8'h80-8 ;
			end
			9'b000000000:
			begin 
				mantissa = 9'b0 ;
				exponent = 8'b0 ;
			end
		endcase	
	end
	
	// get the sign of the output
	assign sign = (int_in < 10'h100)? 1'b1 : 1'b0 ;
	// build output vector
	assign fp_out = {sign, exponent, mantissa} ;
	
endmodule
///////////////////////////////////////////////////////////
// float to 10-bit int 
///////////////////////////////////////////////////////////
// The 50 MHz VGA DAC uses 10-bit binary with 
// zero volts mapped to digital zero
// and about 1.5 volts mapped to digital 1024
// For the tables below, the 'zero' of the sine wave
// is mapped to 10'h100
///////////////////////////////////////////////////////////
module fp2int (int_out, fp_in);
	output [9:0] int_out ;
	input [17:0] fp_in ;
	
	wire sign ;
	wire [7:0] exponent ;
	wire [8:0] mantissa, mtemp ;
	wire [8:0] out_temp ;
	
	assign mantissa = fp_in[8:0];
	assign exponent = fp_in[16:9];
	assign sign = fp_in[17];
	
	assign out_temp = mantissa>>(8'h80-exponent) ;
	assign int_out = sign? (9'h100 - out_temp) : (out_temp + 9'h100) ;
endmodule

//////////////////////////////////////////////////////////
// floating point multiply 
// -- sign bit -- 8-bit exponent -- 9-bit mantissa
// Similar to fp_mult from altera
// NO denorms, no flags, no NAN, no infinity!
//////////////////////////////////////////////////////////
// f1 = {s1, e1, m1), f2 = {s2, e2, m2)
// If either is zero (zero MSB of mantissa) then output is zero
// If e1+e2<129 the result is zero (underflow)
///////////////////////////////////////////////////////////	
module fpmult (fout, f1, f2);

	input [17:0] f1, f2 ;
	output [17:0] fout ;
	
	reg  [17:0] fout ;
	reg sout ;
	reg [8:0] mout ;
	reg [8:0] eout ; // 9-bits for overflow
	
	wire s1, s2;
	wire [8:0] m1, m2 ;
	wire [8:0] e1, e2 ; // extend to 9 bits to avoid overflow
	wire [17:0] mult_out ;	// raw multiplier output
	
	// parse f1
	assign s1 = f1[17]; 	// sign
	assign e1 = {1'b0, f1[16:9]};	// exponent
	assign m1 = f1[8:0] ;	// mantissa
	// parse f2
	assign s2 = f2[17];
	assign e2 = {1'b0, f2[16:9]};
	assign m2 = f2[8:0] ;
	
	// build output
	unsigned_mult mm(mult_out, m1, m2);
	always @(*)
	begin
		if ((f1[8]==0) | (f2[8]==0) | (e1+e2<9'h81)) fout = 18'h0 ;
		else // both inputs are not zero and no exponent underflow
		begin
			sout = s1 ^ s2 ;
			if (mult_out[17]==1)
			begin // MSB of product==1 implies normalized
				eout = e1+e2-9'h80;
				mout = mult_out[17:9] ;
			end
			else // MSB of product==0 implies not normalized
			begin
				eout = e1+e2-9'h81;
				mout = mult_out[16:8] ;
			end
			fout = {sout, eout[7:0], mout} ;	
		end
	end // always @(*)
	
endmodule

///////////////////////////////////////
// low level integer multiply
///////////////////////////////////////
module unsigned_mult (out, a, b);
	output [17:0] out;
	input [8:0] a;
	input [8:0] b;
	assign out = a * b;
endmodule


/////////////////////////////////////////////////////////////////////////////
// floating point Add 
// -- sign bit -- 8-bit exponent -- 9-bit mantissa
// NO denorms, no flags, no NAN, no infinity!
/////////////////////////////////////////////////////////////////////////////
// f1 = {s1, e1, m1), f2 = {s2, e2, m2)
// If either input is zero (zero MSB of mantissa) then output is the remaining input.
// If either input is <(other input)/2**9 then output is the remaining input.
// Sign of the output is the sign of the greater magnitude input
// Add the two inputs if their signs are the same. 
// Subtract the two inputs (bigger-smaller) if their signs are different
//////////////////////////////////////////////////////////////////////////////	
module fpadd (fout, f1, f2);

	input [17:0] f1, f2 ;
	output [17:0] fout ;
	
	reg  [17:0] fout ;
	reg sout ;
	reg [8:0] mout ;
	reg [7:0] eout ;
	reg [18:0] shift_small, denorm_mout ;
	
	wire s1, s2 ; // input signs
	reg  sb, ss ; // signs of bigger and smaller
	wire [8:0] m1, m2 ; // input mantissas
	reg  [8:0] mb, ms ; // mantissas of bigger and smaller
	wire [7:0] e1, e2 ; // input exp
	wire [7:0] ediff ;  // exponent difference
	reg  [7:0] eb, es ; // exp of bigger and smaller
	
	// parse f1
	assign s1 = f1[17]; 	// sign
	assign e1 = f1[16:9];	// exponent
	assign m1 = f1[8:0] ;	// mantissa
	// parse f2
	assign s2 = f2[17];
	assign e2 = f2[16:9];
	assign m2 = f2[8:0] ;
	
	// find biggest magnitude
	always @(*)
	begin
		if (e1>e2) // f1 is bigger
		begin
			sb = s1 ; // the bigger number (absolute value)
			eb = e1 ;
			mb = m1 ;
			ss = s2 ; // the smaller number
			es = e2 ;
			ms = m2 ;
		end
		else if (e2>e1) //f2 is bigger
		begin
			sb = s2 ; // the bigger number (absolute value)
			eb = e2 ;
			mb = m2 ;
			ss = s1 ; // the smaller number
			es = e1 ;
			ms = m1 ;
		end
		else // e1==e2, so need to look at mantissa to determine bigger/smaller
		begin
			if (m1>m2) // f1 is bigger
			begin
				sb = s1 ; // the bigger number (absolute value)
				eb = e1 ;
				mb = m1 ;
				ss = s2 ; // the smaller number
				es = e2 ;
				ms = m2 ;
			end
			else // f2 is bigger or same size
			begin
				sb = s2 ; // the bigger number (absolute value)
				eb = e2 ;
				mb = m2 ;
				ss = s1 ; // the smaller number
				es = e1 ;
				ms = m1 ;
			end
		end
	end //found the bigger number
	
	// do the actual add:
	// -- equalize exponents
	// -- add/sub
	// -- normalize
	assign ediff = eb - es ; // the actual difference in exponents
	always @(*)
	begin
		if ((ms[8]==0) & (mb[8]==0)) fout = 18'h0 ; // both inputs are zero
		else if ((ms[8]==0) | ediff>8) fout = {sb,eb,mb} ; // smaller is too small to matter
		else  // shift/add/normalize
		begin
			sout = sb ;
			// now shift but save the low order bits by extending the registers
			// need a high order bit for 1.0<sum<2.0
			shift_small = {1'b0, ms, 9'b0} >> ediff ;
			// same signs means add -- different means subtract
			if (sb==ss) //do the add
			begin
				denorm_mout = {1'b0, mb, 9'b0} + shift_small ;
				// normalize --
				// when adding result has to be 0.5<sum<2.0
				if (denorm_mout[18]==1) // sum bigger than 1
				begin
					mout = denorm_mout[18:10] ; // take the top bits (shift-right)
					eout = eb + 1 ; // compensate for the shift-right
				end
				else //0.5<sum<1.0
				begin
					mout = denorm_mout[17:9] ; // drop the top bit (no-shift-right)
					eout = eb ; // 
				end
			end // end add logic
			else // otherwise subtract
			begin
				denorm_mout = {1'b0, mb, 9'b0} - shift_small ;
				// the denorm_mout is always smaller then the bigger input
				// (and never an overflow, so bit 18 is always zero)
				// and can be as low as zero! Thus up to 8 shifts may be necessary
				// to normalize denorm_mout
				/* if (denorm_mout[17:9]<9'b1) // is it zero?
				begin
					mout = 9'b0 ;
					eout = 8'b0 ;
				end
				else if (denorm_mout[17:9]<9'b000000010) // shift 8
				begin
					mout = denorm_mout[17:9]<<8 ;
					eout = eb-8 ;
				end
				else if (denorm_mout[17:9]<9'b000000100) // shift 7
				begin
					mout = denorm_mout[17:9]<<7 ;
					eout = eb-7 ;
				end
				else if (denorm_mout[17:9]<9'b000001000) // shift 6
				begin
					mout = denorm_mout[17:9]<<6 ;
					eout = eb-6 ;
				end
				else if (denorm_mout[17:9]<9'b000010000) // shift 5
				begin
					mout = denorm_mout[17:9]<<5 ;
					eout = eb-5 ;
				end
				else if (denorm_mout[17:9]<9'b000100000) // shift 4
				begin
					mout = denorm_mout[17:9]<<4 ;
					eout = eb-4 ;
				end
				else if (denorm_mout[17:9]<9'b001000000) // shift 3
				begin
					mout = denorm_mout[17:9]<<3 ;
					eout = eb-3 ;
				end
				else if (denorm_mout[17:9]<9'b010000000) // shift 2
				begin
					mout = denorm_mout[17:9]<<2 ;
					eout = eb-2 ;
				end
				else if (denorm_mout[17:9]<9'b100000000) // shift 1
				begin
					mout = denorm_mout[17:9]<<1 ;
					eout = eb-1 ;
				end
				else  // shift 0
				begin
					mout = denorm_mout[17:9] ;
					eout = eb ;
				end // end subtract logic */
				
				// normalize denorm_mout
				casex (denorm_mout[17:9])
					9'b1xxxxxxxx: 
					begin 
						mout = denorm_mout[17:9] ;
						eout = eb ;
					end 
					9'b01xxxxxxx: 
					begin 
						mout = denorm_mout[17:9]<<1 ;
						eout = eb-1 ;
					end 
					9'b001xxxxxx: 
					begin 
						mout = denorm_mout[17:9]<<2 ;
						eout = eb-2 ;
					end 
					9'b0001xxxxx: 
					begin 
						mout = denorm_mout[17:9]<<3 ;
						eout = eb-3 ;
					end 
					9'b00001xxxx: 
					begin 
						mout = denorm_mout[17:9]<<4 ;
						eout = eb-4 ;
					end 
					9'b000001xxx: 
					begin 
						mout = denorm_mout[17:9]<<5 ;
						eout = eb-5 ;
					end 
					9'b0000001xx:
 					begin 
						mout = denorm_mout[17:9]<<6 ;
						eout = eb-6 ;
					end 
					9'b00000001x:
 					begin 
						mout = denorm_mout[17:9]<<7 ;
						eout = eb-7 ;
					end 
					9'b000000001:
 					begin 
						mout = denorm_mout[17:9]<<8 ;
						eout = eb-8 ;
					end
					9'b000000000:
 					begin 
						mout = 9'b0 ;
						eout = 8'b0 ;
					end
				endcase	
				
			end // end add/subtract logic
			// format the output
			fout = {sout, eout, mout} ;	
		end // end shift/add/normailize/
	end // always @(*) to compute sum
	
endmodule 


///////////////////////////////////////////////
// Infer ROM storage
// for a sin table for the DDS

module sync_rom (clock, address, sine);
	input clock;
	input [7:0] address;
	output [9:0] sine;
	reg [9:0] sine;
	always @ (posedge clock)
	begin
		case (address)
			8'h00: sine = 10'h100 ;
			8'h01: sine = 10'h106 ;
			8'h02: sine = 10'h10c ;
			8'h03: sine = 10'h112 ;
			8'h04: sine = 10'h118 ;
			8'h05: sine = 10'h11f ;
			8'h06: sine = 10'h125 ;
			8'h07: sine = 10'h12b ;
			8'h08: sine = 10'h131 ;
			8'h09: sine = 10'h137 ;
			8'h0a: sine = 10'h13d ;
			8'h0b: sine = 10'h144 ;
			8'h0c: sine = 10'h14a ;
			8'h0d: sine = 10'h14f ;
			8'h0e: sine = 10'h155 ;
			8'h0f: sine = 10'h15b ;
			8'h10: sine = 10'h161 ;
			8'h11: sine = 10'h167 ;
			8'h12: sine = 10'h16d ;
			8'h13: sine = 10'h172 ;
			8'h14: sine = 10'h178 ;
			8'h15: sine = 10'h17d ;
			8'h16: sine = 10'h183 ;
			8'h17: sine = 10'h188 ;
			8'h18: sine = 10'h18d ;
			8'h19: sine = 10'h192 ;
			8'h1a: sine = 10'h197 ;
			8'h1b: sine = 10'h19c ;
			8'h1c: sine = 10'h1a1 ;
			8'h1d: sine = 10'h1a6 ;
			8'h1e: sine = 10'h1ab ;
			8'h1f: sine = 10'h1af ;
			8'h20: sine = 10'h1b4 ;
			8'h21: sine = 10'h1b8 ;
			8'h22: sine = 10'h1bc ;
			8'h23: sine = 10'h1c1 ;
			8'h24: sine = 10'h1c5 ;
			8'h25: sine = 10'h1c9 ;
			8'h26: sine = 10'h1cc ;
			8'h27: sine = 10'h1d0 ;
			8'h28: sine = 10'h1d4 ;
			8'h29: sine = 10'h1d7 ;
			8'h2a: sine = 10'h1da ;
			8'h2b: sine = 10'h1dd ;
			8'h2c: sine = 10'h1e0 ;
			8'h2d: sine = 10'h1e3 ;
			8'h2e: sine = 10'h1e6 ;
			8'h2f: sine = 10'h1e9 ;
			8'h30: sine = 10'h1eb ;
			8'h31: sine = 10'h1ed ;
			8'h32: sine = 10'h1f0 ;
			8'h33: sine = 10'h1f2 ;
			8'h34: sine = 10'h1f4 ;
			8'h35: sine = 10'h1f5 ;
			8'h36: sine = 10'h1f7 ;
			8'h37: sine = 10'h1f8 ;
			8'h38: sine = 10'h1fa ;
			8'h39: sine = 10'h1fb ;
			8'h3a: sine = 10'h1fc ;
			8'h3b: sine = 10'h1fd ;
			8'h3c: sine = 10'h1fd ;
			8'h3d: sine = 10'h1fe ;
			8'h3e: sine = 10'h1fe ;
			8'h3f: sine = 10'h1fe ;
			8'h40: sine = 10'h1ff ;
			8'h41: sine = 10'h1fe ;
			8'h42: sine = 10'h1fe ;
			8'h43: sine = 10'h1fe ;
			8'h44: sine = 10'h1fd ;
			8'h45: sine = 10'h1fd ;
			8'h46: sine = 10'h1fc ;
			8'h47: sine = 10'h1fb ;
			8'h48: sine = 10'h1fa ;
			8'h49: sine = 10'h1f8 ;
			8'h4a: sine = 10'h1f7 ;
			8'h4b: sine = 10'h1f5 ;
			8'h4c: sine = 10'h1f4 ;
			8'h4d: sine = 10'h1f2 ;
			8'h4e: sine = 10'h1f0 ;
			8'h4f: sine = 10'h1ed ;
			8'h50: sine = 10'h1eb ;
			8'h51: sine = 10'h1e9 ;
			8'h52: sine = 10'h1e6 ;
			8'h53: sine = 10'h1e3 ;
			8'h54: sine = 10'h1e0 ;
			8'h55: sine = 10'h1dd ;
			8'h56: sine = 10'h1da ;
			8'h57: sine = 10'h1d7 ;
			8'h58: sine = 10'h1d4 ;
			8'h59: sine = 10'h1d0 ;
			8'h5a: sine = 10'h1cc ;
			8'h5b: sine = 10'h1c9 ;
			8'h5c: sine = 10'h1c5 ;
			8'h5d: sine = 10'h1c1 ;
			8'h5e: sine = 10'h1bc ;
			8'h5f: sine = 10'h1b8 ;
			8'h60: sine = 10'h1b4 ;
			8'h61: sine = 10'h1af ;
			8'h62: sine = 10'h1ab ;
			8'h63: sine = 10'h1a6 ;
			8'h64: sine = 10'h1a1 ;
			8'h65: sine = 10'h19c ;
			8'h66: sine = 10'h197 ;
			8'h67: sine = 10'h192 ;
			8'h68: sine = 10'h18d ;
			8'h69: sine = 10'h188 ;
			8'h6a: sine = 10'h183 ;
			8'h6b: sine = 10'h17d ;
			8'h6c: sine = 10'h178 ;
			8'h6d: sine = 10'h172 ;
			8'h6e: sine = 10'h16d ;
			8'h6f: sine = 10'h167 ;
			8'h70: sine = 10'h161 ;
			8'h71: sine = 10'h15b ;
			8'h72: sine = 10'h155 ;
			8'h73: sine = 10'h14f ;
			8'h74: sine = 10'h14a ;
			8'h75: sine = 10'h144 ;
			8'h76: sine = 10'h13d ;
			8'h77: sine = 10'h137 ;
			8'h78: sine = 10'h131 ;
			8'h79: sine = 10'h12b ;
			8'h7a: sine = 10'h125 ;
			8'h7b: sine = 10'h11f ;
			8'h7c: sine = 10'h118 ;
			8'h7d: sine = 10'h112 ;
			8'h7e: sine = 10'h10c ;
			8'h7f: sine = 10'h106 ;
			8'h80: sine = 10'h100 ;
			8'h81: sine = 10'h0f9 ;
			8'h82: sine = 10'h0f3 ;
			8'h83: sine = 10'h0ed ;
			8'h84: sine = 10'h0e7 ;
			8'h85: sine = 10'h0e0 ;
			8'h86: sine = 10'h0da ;
			8'h87: sine = 10'h0d4 ;
			8'h88: sine = 10'h0ce ;
			8'h89: sine = 10'h0c8 ;
			8'h8a: sine = 10'h0c2 ;
			8'h8b: sine = 10'h0bb ;
			8'h8c: sine = 10'h0b5 ;
			8'h8d: sine = 10'h0b0 ;
			8'h8e: sine = 10'h0aa ;
			8'h8f: sine = 10'h0a4 ;
			8'h90: sine = 10'h09e ;
			8'h91: sine = 10'h098 ;
			8'h92: sine = 10'h092 ;
			8'h93: sine = 10'h08d ;
			8'h94: sine = 10'h087 ;
			8'h95: sine = 10'h082 ;
			8'h96: sine = 10'h07c ;
			8'h97: sine = 10'h077 ;
			8'h98: sine = 10'h072 ;
			8'h99: sine = 10'h06d ;
			8'h9a: sine = 10'h068 ;
			8'h9b: sine = 10'h063 ;
			8'h9c: sine = 10'h05e ;
			8'h9d: sine = 10'h059 ;
			8'h9e: sine = 10'h054 ;
			8'h9f: sine = 10'h050 ;
			8'ha0: sine = 10'h04b ;
			8'ha1: sine = 10'h047 ;
			8'ha2: sine = 10'h043 ;
			8'ha3: sine = 10'h03e ;
			8'ha4: sine = 10'h03a ;
			8'ha5: sine = 10'h036 ;
			8'ha6: sine = 10'h033 ;
			8'ha7: sine = 10'h02f ;
			8'ha8: sine = 10'h02b ;
			8'ha9: sine = 10'h028 ;
			8'haa: sine = 10'h025 ;
			8'hab: sine = 10'h022 ;
			8'hac: sine = 10'h01f ;
			8'had: sine = 10'h01c ;
			8'hae: sine = 10'h019 ;
			8'haf: sine = 10'h016 ;
			8'hb0: sine = 10'h014 ;
			8'hb1: sine = 10'h012 ;
			8'hb2: sine = 10'h00f ;
			8'hb3: sine = 10'h00d ;
			8'hb4: sine = 10'h00b ;
			8'hb5: sine = 10'h00a ;
			8'hb6: sine = 10'h008 ;
			8'hb7: sine = 10'h007 ;
			8'hb8: sine = 10'h005 ;
			8'hb9: sine = 10'h004 ;
			8'hba: sine = 10'h003 ;
			8'hbb: sine = 10'h002 ;
			8'hbc: sine = 10'h002 ;
			8'hbd: sine = 10'h001 ;
			8'hbe: sine = 10'h001 ;
			8'hbf: sine = 10'h001 ;
			8'hc0: sine = 10'h001 ;
			8'hc1: sine = 10'h001 ;
			8'hc2: sine = 10'h001 ;
			8'hc3: sine = 10'h001 ;
			8'hc4: sine = 10'h002 ;
			8'hc5: sine = 10'h002 ;
			8'hc6: sine = 10'h003 ;
			8'hc7: sine = 10'h004 ;
			8'hc8: sine = 10'h005 ;
			8'hc9: sine = 10'h007 ;
			8'hca: sine = 10'h008 ;
			8'hcb: sine = 10'h00a ;
			8'hcc: sine = 10'h00b ;
			8'hcd: sine = 10'h00d ;
			8'hce: sine = 10'h00f ;
			8'hcf: sine = 10'h012 ;
			8'hd0: sine = 10'h014 ;
			8'hd1: sine = 10'h016 ;
			8'hd2: sine = 10'h019 ;
			8'hd3: sine = 10'h01c ;
			8'hd4: sine = 10'h01f ;
			8'hd5: sine = 10'h022 ;
			8'hd6: sine = 10'h025 ;
			8'hd7: sine = 10'h028 ;
			8'hd8: sine = 10'h02b ;
			8'hd9: sine = 10'h02f ;
			8'hda: sine = 10'h033 ;
			8'hdb: sine = 10'h036 ;
			8'hdc: sine = 10'h03a ;
			8'hdd: sine = 10'h03e ;
			8'hde: sine = 10'h043 ;
			8'hdf: sine = 10'h047 ;
			8'he0: sine = 10'h04b ;
			8'he1: sine = 10'h050 ;
			8'he2: sine = 10'h054 ;
			8'he3: sine = 10'h059 ;
			8'he4: sine = 10'h05e ;
			8'he5: sine = 10'h063 ;
			8'he6: sine = 10'h068 ;
			8'he7: sine = 10'h06d ;
			8'he8: sine = 10'h072 ;
			8'he9: sine = 10'h077 ;
			8'hea: sine = 10'h07c ;
			8'heb: sine = 10'h082 ;
			8'hec: sine = 10'h087 ;
			8'hed: sine = 10'h08d ;
			8'hee: sine = 10'h092 ;
			8'hef: sine = 10'h098 ;
			8'hf0: sine = 10'h09e ;
			8'hf1: sine = 10'h0a4 ;
			8'hf2: sine = 10'h0aa ;
			8'hf3: sine = 10'h0b0 ;
			8'hf4: sine = 10'h0b5 ;
			8'hf5: sine = 10'h0bb ;
			8'hf6: sine = 10'h0c2 ;
			8'hf7: sine = 10'h0c8 ;
			8'hf8: sine = 10'h0ce ;
			8'hf9: sine = 10'h0d4 ;
			8'hfa: sine = 10'h0da ;
			8'hfb: sine = 10'h0e0 ;
			8'hfc: sine = 10'h0e7 ;
			8'hfd: sine = 10'h0ed ;
			8'hfe: sine = 10'h0f3 ;
			8'hff: sine = 10'h0f9 ;
		endcase
	end
endmodule
///////////////////////////////////////////////